// Constants used to configure the camera and the layouts
var LAYOUT_DURATION = 400;
var LOCATE_OPTIONS = {
  duration: 400,
  padding: { top: 200, bottom: 80, left: 50, right: 50 }
};

// Create an instance of Ogma
// We know we are dealing with a small graph, so we can afford to use the canvas renderer
var ogma = new Ogma({ renderer: 'canvas' });

// Retrieve the fake database defined in `scripts/dummyDatabase.js`
var DB = window.DB;

// Retrieve the important elements of the UI
var searchBar = document.getElementById('search-bar');
var boxHideEvaluators = document.getElementById('box-hide-evaluators');
var boxHideSmallClaims = document.getElementById('box-hide-small-claims');
var boxHideLeaves = document.getElementById('box-hide-leaves');
var boxTexts = document.getElementById('box-assign-texts');
var boxIcons = document.getElementById('box-assign-icons');
var boxColors = document.getElementById('box-assign-colors');
var boxClaims = document.getElementById('box-assign-claim-sizes');
var boxShowLegend = document.getElementById('box-show-legend');
var buttonForceLink = document.getElementById('btn-forceLink');
var buttonHierarchical = document.getElementById('btn-hierarchical');
var buttonSearch = document.getElementById('btn-search');
var buttonExport = document.getElementById('btn-export');
var buttonCenter = document.getElementById('btn-center');
var buttonZoomOut = document.getElementById('btn-zoom-out');
var buttonZoomIn = document.getElementById('btn-zoom-in');
var buttonReset = document.getElementById('btn-reset');
var buttonDisplayAll = document.getElementById('btn-display-all');

// Initialize the style rules/filters
var textRule = null;
var iconRule = null;
var colorRule = null;
var sizeRule = null;
var evaluatorFilter = null;
var smallClaimFilter = null;
var leafFilter = null;

// Main function, initializes the visualization
function init() {
  // Assign the HTML container to Ogma, and set the background color to be a bit darker that white
  ogma.setContainer('graph-container');
  ogma.setOptions({ backgroundColor: '#F0F0F0' });

  // Bind the buttons/checkboxes to the associated actions
  boxHideEvaluators.addEventListener('click', applyEvaluatorFilter);
  boxHideSmallClaims.addEventListener('click', applySmallClaimFilter);
  boxHideLeaves.addEventListener('click', applyLeafFilter);
  boxTexts.addEventListener('click', applyTextRule);
  boxColors.addEventListener('click', applyColorRule);
  boxIcons.addEventListener('click', applyIconRule);
  boxClaims.addEventListener('click', applyClaimSizeRule);
  boxShowLegend.addEventListener('click', toggleLegend);
  buttonForceLink.addEventListener('click', runForceLink);
  buttonHierarchical.addEventListener('click', runHierarchical);
  buttonZoomIn.addEventListener('click', zoomIn);
  buttonZoomOut.addEventListener('click', zoomOut);
  buttonCenter.addEventListener('click', centerView);
  buttonSearch.addEventListener('click', searchNode);
  buttonExport.addEventListener('click', exportToPng);
  buttonReset.addEventListener('click', reset);
  buttonDisplayAll.addEventListener('click', displayAll);

  // Trigger the search on "Enter" when search bar has focus
  document.addEventListener('keydown', function(evt) {
    var code = evt.keyCode || evt.which;
    if (code === 13 && document.activeElement === searchBar) {
      searchNode();
    }
  });

  // Expand a node when double-clicked
  ogma.events.onDoubleClick(function(evt) {
    if (evt.target && evt.target.isNode && evt.button === 'left') {
      expandNeighbors(evt.target);

      // Clicking on a node adds it to the selection, but we don't want a node to
      // be selected when double-clicked
      evt.target.setSelected(!evt.target.isSelected());
    }
  });

  // Styling rules that will be applied globally
  ogma.styles.addNodeRule({
    text: {
      size: 14
    },
    icon: {
      font: 'Font Awesome 5 Free',
      style: 'bold'
    },
    badges: {
      bottomRight: {
        color: 'inherit',
        text: {
          scale: 0.8,
          content: function(node) {
            // The bottom right badge displays the number of hidden neighbors
            // The `nbNeighbors` data property contains the total number of neighbors for the node in the DB
            // The `getDegree` method retrieves the number of neighbors displayed in the viz
            return node.getData('nbNeighbors') - node.getDegree() || null;
          }
        }
      }
    }
  });

  ogma.styles.addEdgeRule({
    shape: 'arrow'
  });

  // Show the type of the edges on hover
  ogma.styles.setHoveredEdgeAttributes({
    text: function(node) {
      return node.getData('type');
    }
  });

  // Display the data of nodes as a tooltip on hover
  ogma.tools.tooltip.onNodeHover(
    function(node) {
      var html = [
        '<div class="arrow"></div>',
        '<div class="ogma-tooltip-header"><span class="icon">' +
          node.getAttribute('icon.content') +
          '</span> ' +
          node.getData('type') +
          '</div>',
        '<div class="ogma-tooltip-body">',
        '<table>'
      ];

      var properties = node.getData('properties');

      Object.keys(properties).forEach(function(key) {
        var value = properties[key];

        html.push('<tr><th>' + key + '</th><td>' + value + '</td></tr>');
      });

      html.push(
        '</table></div><div class="ogma-tooltip-footer">Number of connections: ' +
          node.getData('nbNeighbors') +
          '</div>'
      );

      return html.join('');
    },
    { className: 'ogma-tooltip' }
  );

  // Set the focus on the search bar
  searchBar.focus();

  // Set the default value for the search bar
  searchBar.value = 'John Piggyback';

  // Set the cursor at the end of the search bar
  searchBar.setSelectionRange(searchBar.value.length, searchBar.value.length);

  boxTexts.checked = true;
  boxColors.checked = true;
  boxIcons.checked = true;

  // Apply the text, color and icon rules
  applyTextRule();
  applyColorRule();
  applyIconRule();
}

// Add the given node to the visualization and returns the added node if it exist
function addNodeToViz(name) {
  // Look for the node in the DB
  var node = DB.search(name);

  if (node) {
    var addedNode = ogma.addNode(node);

    if (ogma.getNodes().size === 1) {
      // If this is the first node in the visualization, we simply center the camera on it
      ogma.view.locateGraph();
    } else {
      // If there are more than one node, apply the forceLink layout (force-directed)
      ogma.addEdges(selectAdjacentEdgesToAdd([node.id])).then(function() {
        runLayout('forceLink');
      });
    }

    return addedNode;
  } else {
    return null;
  }
}

// Retrieve the list of adjacent edges to the specified nodes, for which both the
// source and target are already loaded in Ogma (in the viz)
function selectAdjacentEdgesToAdd(nodeIds) {
  return DB.getAdjacentEdges(nodeIds).filter(function(edge) {
    return ogma.getNode(edge.source) && ogma.getNode(edge.target);
  });
}

// Expand the specified node by retrieving its neighbors from the database
// and adding them to the visualization
function expandNeighbors(node) {
  // Retrieve the neighbors from the DB
  var neighbors = DB.getNeighbors(node.getId()),
    ids = neighbors.nodeIds,
    nodes = neighbors.nodes;

  // If this condition is false, it means that all the retrieved nodes are already in Ogma.
  // In this case we do nothing
  if (ogma.getNodes(ids).size < ids.length) {
    // Set the position of the neighbors around the nodes, in preparation to the force-directed layout
    var position = node.getPosition(),
      angleStep = (2 * Math.PI) / neighbors.nodes.length,
      angle = Math.random() * angleStep;

    for (var i = 0; i < nodes.length; ++i) {
      var neighbor = nodes[i];
      neighbor.attributes = {
        x: position.x + Math.cos(angle) * 0.001,
        y: position.y + Math.sin(angle) * 0.001
      };

      angle += angleStep;
    }

    // Add the neighbors to the visualization, add their adjacent edges and run a layout
    ogma
      .addNodes(nodes)
      .then(function() {
        return ogma.addEdges(selectAdjacentEdgesToAdd(ids));
      })
      .then(function() {
        return runLayout('forceLink');
      });
  }
}

/* ============================== */
/* Function triggered by the menu */
/* ============================== */

function applyEvaluatorFilter() {
  if (boxHideEvaluators.checked && !evaluatorFilter) {
    evaluatorFilter = ogma.addNodeFilter(function(node) {
      return node.getData('type') !== 'Evaluator';
    });
  } else if (!boxHideEvaluators.checked && evaluatorFilter) {
    evaluatorFilter.delete();
    evaluatorFilter = null;
  }
}

function applySmallClaimFilter() {
  if (boxHideSmallClaims.checked && !smallClaimFilter) {
    smallClaimFilter = ogma.addNodeFilter(function(node) {
      return (
        node.getData('type') !== 'Claim' ||
        node.getData('properties.amount') >= 50000
      );
    });
  } else if (!boxHideSmallClaims.checked && smallClaimFilter) {
    smallClaimFilter.delete();
    smallClaimFilter = null;
  }
}

function applyLeafFilter() {
  if (boxHideLeaves.checked && !leafFilter) {
    leafFilter = ogma.addNodeFilter(function(node) {
      return node.getAdjacentNodes().size > 1;
    });
  } else if (!boxHideLeaves.checked && leafFilter) {
    leafFilter.delete();
    leafFilter = null;
  }
}

function applyTextRule() {
  if (boxTexts.checked && !textRule) {
    textRule = ogma.styles.addNodeRule({
      text: {
        content: function(node) {
          var type = node.getData('type');

          if (
            type === 'Customer' ||
            type === 'Lawyer' ||
            type === 'Evaluator'
          ) {
            return node.getData('properties.fullname');
          } else if (
            type === 'Phone' ||
            type === 'MailAddress' ||
            type === 'SSN'
          ) {
            return node.getData('properties.name');
          } else if (type === 'Address') {
            return (
              node.getData('properties.city') +
              ', ' +
              node.getData('properties.state')
            );
          } else if (type === 'Claim') {
            return (
              node.getData('properties.name') +
              ' (' +
              node.getData('properties.amount') +
              '$)'
            );
          }
        }
      }
    });
  } else if (!boxTexts.checked && textRule) {
    textRule.delete();
    textRule = null;
  }
}

function applyColorRule() {
  if (boxColors.checked && !colorRule) {
    colorRule = ogma.styles.addNodeRule({
      color: ogma.rules.map({
        field: 'type',
        values: {
          Customer: '#FF5217',
          Claim: '#E8AB15',
          Address: '#D0FF24',
          SSN: '#0FE85C',
          Phone: '#0EB1FF',
          MailAddress: '#FF1271',
          Lawyer: '#AE15FF',
          Evaluator: '#1067FF'
        }
      })
    });
  } else if (!boxColors.checked && colorRule) {
    colorRule.delete();
    colorRule = null;
  }
}

function applyIconRule() {
  if (boxIcons.checked && !iconRule) {
    iconRule = ogma.styles.addNodeRule({
      icon: {
        content: ogma.rules.map({
          field: 'type',
          values: {
            Customer: '\uf007',
            Claim: '\uf155',
            Address: '\uf015',
            SSN: '\uf2c2',
            Phone: '\uf095',
            MailAddress: '\uf1fa',
            Lawyer: '\uf0e3',
            Evaluator: '\uf508'
          }
        })
      }
    });
  } else if (!boxIcons.checked && iconRule) {
    iconRule.delete();
    iconRule = null;
  }
}

function applyClaimSizeRule() {
  if (boxClaims.checked && !sizeRule) {
    sizeRule = ogma.styles.addNodeRule(
      function(node) {
        return node.getData('type') === 'Claim';
      },
      {
        radius: ogma.rules.slices({
          field: 'properties.amount',
          values: { min: 3, max: 7 },
          stops: { min: 48000, max: 52000 }
        })
      }
    );
  } else if (!boxClaims.checked && sizeRule) {
    sizeRule.delete();
    sizeRule = null;
  }
}

function toggleLegend() {
  if (boxShowLegend.checked) {
    ogma.tools.legend.enable({
      fontSize: 11,
      titleFunction: function(propertyPath) {
        var propName = propertyPath[propertyPath.length - 1];

        if (propName === 'amount') {
          return 'Claim amount';
        } else {
          return propName.charAt(0).toUpperCase() + propName.substring(1);
        }
      }
    });
  } else {
    ogma.tools.legend.disable();
  }
}

// Utility function to run a layout
function runLayout(name) {
  return ogma.layouts[name]({
    locate: LOCATE_OPTIONS,
    duration: LAYOUT_DURATION
  });
}

function runForceLink() {
  runLayout('forceLink');
}

function runHierarchical() {
  runLayout('hierarchical');
}

function displayAll() {
  // Retrieve the whole graph from the database
  var graph = DB.getFullGraph();

  // If there is the same amount of nodes in Ogma as in the DB, nothing to do
  if (ogma.getNodes('all').size !== graph.nodes.length) {
    // Assign a random position to all nodes, in preparation for the layout
    for (var i = 0; i < graph.nodes.length; ++i) {
      graph.nodes[i].attributes = {
        x: Math.random(),
        y: Math.random()
      };
    }

    ogma
      .setGraph(graph)
      .then(function() {
        return ogma.view.locateGraph(LOCATE_OPTIONS);
      })
      .then(function() {
        runForceLink();
      });
  }
}

function reset() {
  ogma.clearGraph();
  boxHideEvaluators.checked = false;
  boxHideSmallClaims.checked = false;
  boxHideLeaves.checked = false;
  boxTexts.checked = true;
  boxColors.checked = true;
  boxIcons.checked = true;
  boxClaims.checked = false;
  boxShowLegend.checked = false;

  applyTextRule();
  applyLeafFilter();
  applySmallClaimFilter();
  applyEvaluatorFilter();
  applyClaimSizeRule();
  applyColorRule();
  applyIconRule();
  toggleLegend();
}

/* ================================================== */
/* Function triggered by buttons in the visualization */
/* ================================================== */

function zoomIn() {
  ogma.view.zoomIn({
    duration: 150,
    easing: 'quadraticOut'
  });
}

function zoomOut() {
  ogma.view.zoomOut({
    duration: 150,
    easing: 'quadraticOut'
  });
}

function searchNode() {
  if (searchBar.value) {
    if (addNodeToViz(searchBar.value)) {
      searchBar.value = '';
    } else {
      alert(
        'No node has the property "fullname" equal to "' +
          searchBar.value +
          '".'
      );
    }
  }
}

function centerView() {
  return ogma.view.locateGraph(LOCATE_OPTIONS);
}

function exportToPng() {
  ogma.export.png({
    background: '#F0F0F0',
    filename: 'Linkurious_Ogma_Library_Example'
  });
}

init();
